(*************************************************************************************************************************
** ExponentialCurveFit.m
** This .m file contains functions to compute the locally adaptive frame (gauge frame) in SE(3)
**
** Author: M.H.J.Janssen <m.h.j.janssen@tue.nl>
** Author: F.C. Martin <f.c.martin@tue.nl>
**************************************************************************************************************************)
BeginPackage["LieAnalysis`OrientationScores3D`GaugeFrames`ExponentialCurveFit`",{
  "LieAnalysis`"
}];


	
	Begin["`Private`"];
	
	
	
		Needs["Classes`"];
		Get["LieAnalysis`Common`"];
		
		
		Options[OrientationScoreGaugeFrames3D] = {
			"TensorType"->"Hessian"
		};
		
		OrientationScoreGaugeFrames3D[ ostObj_ /; ValidQ[ostObj, LieAnalysis`Objects`Obj3DPositionOrientationData`Obj3DPositionOrientationData], {sigmaSpatial_, sigmaOrientationIn_,sigmaSpatialExternal_:0,sigmaAngularExternal_:0}, opts:OptionsPattern[] ] := Module[
	
			{
				orientationList = ostObj["FullOrientationList"],
				msg,spatialTangent, frame, frameNew, cFinal, gaugeFrame, printTmp, beta, gamma, Xi, horizontalTangent, interpolationFunctions, betaNew, gammaNew, spatialComponents, horizontalComponents, orientationListSpatialFit, sigmaOrientation
			},
			
			(* Scaling Between orientation and spatial domain *)
			Xi = sigmaOrientation/sigmaSpatial;
			
			sigmaOrientation = If[sigmaOrientationIn === Automatic,
				 (0.216667 + 1.06061 ostObj[["Wavelets"]][["SigmaAngle"]]),
				 sigmaOrientationIn
			];
			
			(* Create Left Invariant Interpolation Functions *)
			interpolationFunctions = CreateInterpolationFunctions[ostObj,sigmaSpatial,sigmaOrientation,OptionValue["TensorType"]];
			
			(* STEP 1: Determine Spatial Tangent *)
			msg = Row[{ProgressIndicator[0.]," Computing Spatial Tangent"}];
			printTmp = If[OptionValue[LieAnalysis,"DynamicUpdates"], PrintTemporary[Dynamic@msg]];
			{beta,gamma} = ostObj["FullPolarOrientationList"];
			
			(* Compute the spatial Components from either the Structure Tensor or Product Hessian *)
			spatialComponents = ComputeSpatialComponents[beta,gamma, Xi, interpolationFunctions]; (*6,x,y,z,R*)	
			
			(* Apply External Regularization on the Spatial Part *)
			If[sigmaSpatialExternal != 0 || sigmaAngularExternal != 0,
				msg = Row[{ProgressIndicator[0.]," Spatial Components External Regulatization"}];
				spatialComponents = SpatialComponentsExternalRegularization[orientationList,spatialComponents,sigmaSpatialExternal,sigmaAngularExternal];(*6,x,y,z,R*)
			];
			
			(* Eigen System Analyse of the Spatial Tensor *)
			msg = Row[{ProgressIndicator[0.]," Eigen Vector Analysis on Spatial Components"}];
			spatialTangent = LieAnalysis`Common`RealSymmetricThreeEigensystem[Transpose[spatialComponents,{5,1,2,3,4}],"Output"->"EIGENVECTORS"][[All,All,All,All,-1]];
			frame = ConstantArray[LieAnalysis`Common`RnWithAlphaIs0[orientationList],Dimensions[ostObj["Data"],3]];
			orientationListSpatialFit =  MapThread[#1.#2&,{frame,spatialTangent},4];
			
			(* STEP 2: Determine Horizontal Tangent *)
			msg = Row[{ProgressIndicator[0.]," Compute Horizontal Components"}];
			{betaNew,gammaNew} =  Transpose[LieAnalysis`Common`CartesianToSphericalCoordinates[orientationListSpatialFit][[All,All,All,All,2;;3]],{2,3,4,5,1}];
			
			horizontalComponents = ComputeHorizontalComponents[betaNew,gammaNew,Xi, interpolationFunctions];
			
			(* Apply External Regularization on the Horizontal Part *)
			If[sigmaSpatialExternal != 0 || sigmaAngularExternal != 0,
				msg = Row[{ProgressIndicator[0.]," Horizontal Components Extrenal Regularization"}];
				additionalComponents = ComputeExtraComponentsForHorizontalExternalRegularization[betaNew,gammaNew,Xi,interpolationFunctions];
				allComponents = Join[additionalComponents,horizontalComponents];
				horizontalComponents = HorizontalComponentsExternalRegularization[orientationList,orientationListSpatialFit,allComponents,sigmaSpatialExternal,sigmaAngularExternal];
			];
			
			(* Eigen System Analysis of the Horizontal Components *)
			msg = Row[{ProgressIndicator[0.]," Eigen Vector Analysis on Horizontal Components"}];
			horizontalTangent = LieAnalysis`Common`RealSymmetricThreeEigensystem[Transpose[horizontalComponents,{5,1,2,3,4}],"Output"->"EIGENVECTORS"][[All,All,All,All,-1]];
			horizontalTangent = horizontalTangent.DiagonalMatrix[{1/Xi,1,1}];
			horizontalTangent = Map[Normalize,horizontalTangent,{4}];
			frameNew = LieAnalysis`Common`RnWithAlphaIs0[orientationListSpatialFit];
			
			(* Compute the final tangent coefficients *)
			msg = Row[{ProgressIndicator[0.75]," Creating cFinal"}];
			cFinal = MapThread[LieAnalysis`Common`BlockMatrix[Transpose[#1].#2,Transpose[#1].#2][[All,3;;5]].#3&,{frame,frameNew,horizontalTangent},4];
			
			(* Construct final frame *)
			msg = Row[{ProgressIndicator[0.90]," Constructing final frame"}];
			gaugeFrame = LieAnalysis`OrientationScores3D`GaugeFrames`ExponentialCurveFit`Private`TangentVectorToGaugeFrame[cFinal, Xi];
			
			(* Return Gauge frame*)
			NotebookDelete[printTmp];
			Return[gaugeFrame];
					
		];
		
		
		(* 
			Structure Tensor Specific Functions 
		*)
		
		CreateInterpolationFunctions[ostObj_,sigmaSpatial_,sigmaOrientation_,"StructureTensor"] := 
			LieAnalysis`OrientationScores3D`Derivatives`LeftInvariantDerivatives`Private`LeftInvariantDerivativesInternal[
				ostObj, sigmaSpatial, sigmaOrientation,{1,2,3,4,5},"SphericalHarmonics",True];
		
		ComputeSpatialComponents[beta_,gamma_, Xi_, {A1f_,A2f_,A3f_,_,_}] := Module[
	
			{
				spatialComponentsOfStructureTensor, A1, A2, A3
			},
			
			(* Evaluate at Specific Orientations *)
			{A1,A2,A3} = #[beta,gamma]&/@{A1f,A2f,A3f};
			
			(* Compute the Spatial part of the Structure Tensor and Scale with Xi *)
			spatialComponentsOfStructureTensor = {A1^2, A1 A2, A1 A3, A2^2, A2 A3 ,A3^2}/Xi^2;
			
			(* Return Spatial Components of the Structure Tensor *)
			Return[spatialComponentsOfStructureTensor]
			
		];
		
		ComputeHorizontalComponents[betaNew_,gammaNew_, Xi_, {_,_,A3f_,A4f_,A5f_}] := Module[
	
			{
				A3,A4,A5,horizontalComponents
			},
			
			(* Evaluate at Specific Orientations *)
			{A3,A4,A5} = #[betaNew,gammaNew]&/@{A3f, A4f, A5f};
			
			(* Compute Horizotal part of the Structure Tensor and Scale with Xi*)
			horizontalComponents = {A3^2, A3 A4, A3 A5,A4^2,A4 A5,A5^2}*{1/Xi^2,1/Xi,1/Xi,1,1,1};
			
			(* Return the Horizontal Components of the Structure Tensor *)
			Return[horizontalComponents];
					
		];
		
		(* 
			For External regularization on the Horizontal Parts we need the Entire Structure Tensor evaluated on 
			betaNew and gammaNew as we rotate it back to the cartesian frame prior apply regularization. Therefor
			We need to compute additional components: A1^2,A1 A2,A1 A3, A1 A4, A1 A5, A2^2, A3 A4, A3 A5, A4^2, A4 A5, A5^2.
		*)
		ComputeExtraComponentsForHorizontalExternalRegularization[betaNew_,gammaNew_,Xi_,{A1f_,A2f_,A3f_,A4f_,A5f_}] := Module[
		
			{
				A1,A2,A3,A4,A5, zeros, missingComponents
			},
			
			(* Evaluate First order Derivatives at Spatial Fitted Orientations *)
			{A1,A2,A3,A4,A5} = #[betaNew,gammaNew]&/@{A1f,A2f,A3f,A4f,A5f};
			
			(* Because A6=0 we simply can make a zero Matrix *)
			zeros = ConstantArray[0,Dimensions[A1]];
			
			(* Construct Components in addition to the Horizontal Components *)
			missingComponents = {
				(*11*) (A1^2) /Xi^2,
				(*12*) (A1 A2) /Xi^2,
				(*13*) (A1 A3) /Xi^2, 
				(*14*) (A1 A4)/Xi, 
				(*15*) (A1 A5)/Xi, 
				(*16*) zeros,
				(*22*) A2^2/Xi^2,
				(*23*) (A2 A3)/Xi^2, 
				(*24*) (A2 A4)/Xi, 
				(*25*) (A2 A5)/Xi,
				(*26*) zeros,
				(*36*) zeros,
				(*46*) zeros,
				(*56*) zeros,
				(*66*) zeros
			};
			
			(* Return the Components *)
			Return[missingComponents];
			
		];
		
		(* 
			Product Hessian Specific Functions 
		*)
		
		CreateInterpolationFunctions[ostObj_,sigmaSpatial_,sigmaOrientation_,"Hessian"] := 
			LieAnalysis`OrientationScores3D`Derivatives`LeftInvariantDerivatives`Private`LeftInvariantDerivativesInternal[
				ostObj, sigmaSpatial, sigmaOrientation,{1,2,3,4,5,11,12,13,14,15,22,23,24,25,33,34,35,44,45,55},"SphericalHarmonics",True];
					
		ComputeSpatialComponents[beta_,gamma_, Xi_, {_,_,_,_,_,A11f_,A12f_,A13f_,A14f_,A15f_,A22f_,A23f_,A24f_,A25f_,A33f_,A34f_,A35f_,_,_,_}] := Module[
			
			{
				spatialPartOfHessianProduct,
				A11,A12,A13,A14,A15,A22,A23,A24,A25,A33,A34,A35
			},
			
			(* Evaluate at Specific Orientations *)
			{A11,A12,A13,A14,A15,A22,A23,A24,A25,A33,A34,A35} 
				= #[beta,gamma]&/@{A11f,A12f,A13f,A14f,A15f,A22f,A23f,A24f,A25f,A33f,A34f,A35f};
			
			(* Compute the spatial part of the hessian and scale it with Xi *)
			spatialPartOfHessianProduct = {
				A11^2/Xi^4		+	A12^2/Xi^4		+	A13^2/Xi^4		+	A14^2/Xi^2		+	A15^2/Xi^2,
				(A11 A12)/Xi^4	+	(A12 A22)/Xi^4	+	(A13 A23)/Xi^4	+	(A14 A24)/Xi^2	+	(A15 A25)/Xi^2,
				(A11 A13)/Xi^4	+	(A12 A23)/Xi^4	+	(A13 A33)/Xi^4	+	(A14 A34)/Xi^2	+	(A15 A35)/Xi^2,
				A12^2/Xi^4		+	A22^2/Xi^4		+	A23^2/Xi^4		+	A24^2/Xi^2		+	A25^2/Xi^2,
				(A12 A13)/Xi^4	+	(A22 A23)/Xi^4	+	(A23 A33)/Xi^4	+	(A24 A34)/Xi^2	+	(A25 A35)/Xi^2,
				A13^2/Xi^4		+	A23^2/Xi^4		+	A33^2/Xi^4		+	A34^2/Xi^2		+	A35^2/Xi^2
			};
			
			(* Return the spatial part of the Hessian Product *)
			Return[spatialPartOfHessianProduct];
			
		];
		
		ComputeHorizontalComponents[betaNew_,gammaNew_, Xi_, {A1f_,A2f_,A3f_,_,_,_,_,A13f_,A14f_,A15f_,_,A23f_,A24f_,A25f_,A33f_,A34f_,A35f_,A44f_,A45f_,A55f_}] := Module[
			
			{
				A34,A35,A44,A45,A55,A1,A2,A13,A23,A14,A24,A3,A15,A25,
				A33,horizontalPartOfHessian
			},
			
			(* Evaluate at specificOrientations orientations *)
			{A33,A34,A35,A44,A45,A55,A1,A2,A13,A23,A14,A24,A3,A15,A25}
				= #[betaNew,gammaNew]&/@{A33f,A34f,A35f,A44f,A45f,A55f,A1f,A2f,A13f,A23f,A14f,A24f,A3f,A15f,A25f};
			
			(* Compute the horizontal part of hessian and scale it with Xi *)
			horizontalPartOfHessian = {
				A13^2/Xi^4		+	A23^2/Xi^4			+	A33^2/Xi^4			+	A34^2/Xi^2				+	A35^2/Xi^2,
				(A13 A14)/Xi^3	+	(A23 (A24+A3))/Xi^3	+	(A33 (-A2+A34))/Xi^3+	(A34 A44)/Xi			+	(A35 A45)/Xi,
				(A23 A25)/Xi^3	+	(A13 (A15-A3))/Xi^3	+	(A33 (A1+A35))/Xi^3	+	(A34 A45)/Xi			+	(A35 A55)/Xi,
				A44^2			+	A45^2				+	A14^2/Xi^2			+	(A24+A3)^2/Xi^2			+	(-A2+A34)^2/Xi^2,
				A44 A45			+	A45 A55				+	(A14 (A15-A3))/Xi^2	+	(A25 (A24+A3))/Xi^2		+	((-A2+A34) (A1+A35))/Xi^2,
				A45^2			+	A55^2				+	A25^2/Xi^2			+	(A15-A3)^2/Xi^2			+	(A1+A35)^2/Xi^2
			};
			
			(* Return the Horizontal Part of the Hessian *)
			Return[horizontalPartOfHessian];
			
		];
		
		(* 
			For External regularization on the Horizontal Parts we need the Entire Hessian evaluated on 
			betaNew and gammaNew as we rotate it back to the cartesian frame prior apply regularization. Therefor
			We need to compute additional components: A1^2,A1 A2,A1 A3, A1 A4, A1 A5, A2^2, A3 A4, A3 A5, A4^2, A4 A5, A5^2.
		*)
		ComputeExtraComponentsForHorizontalExternalRegularization[betaNew_,gammaNew_,Xi_,{A1f_,A2f_,A3f_,A4f_,A5f_,A11f_,A12f_,A13f_,A14f_,A15f_,A22f_,A23f_,A24f_,A25f_,A33f_,A34f_,A35f_,A44f_,A45f_,A55f_}] := Module[
		
			{
				A1,A2,A3,A4,A5,A11,A12,A13,A14,A15,A22,A23,A24,A25,A33,A34,A35,A44,A45,A55,missingComponents
			},
			
			(* Evaluate at Specific Orientations *)
			{A1,A2,A3,A4,A5,A11,A12,A13,A14,A15,A22,A23,A24,A25,A33,A34,A35,A44,A45,A55} = #[betaNew,gammaNew]&/@{A1f,A2f,A3f,A4f,A5f,A11f,A12f,A13f,A14f,A15f,A22f,A23f,A24f,A25f,A33f,A34f,A35f,A44f,A45f,A55f};
			
			(* Compute Missing Components to create a full Hessian at specific betas and gammas*)
			missingComponents = {
				(*11*)A11^2/Xi^4 + A12^2/Xi^4 + A13^2/Xi^4 + A14^2/Xi^2 + A15^2/Xi^2, (*B11^2+B12^2+B13^2+B14^2+B15^2*)
				(*12*)(A11 A12)/Xi^4 + (A12 A22)/Xi^4 + (A13 A23)/Xi^4 + (A14 A24)/Xi^2 + (A15 A25)/Xi^2, (* B11 B21+B12 B22+B13 B23+B14 B24+B15 B25 *)
				(*13*)(A11 A13)/Xi^4 + (A12 A23)/Xi^4 + (A13 A33)/Xi^4 + (A14 A34)/Xi^2 + (A15 A35)/Xi^2, (*B11 B31+B12 B32+B13 B33+B14 B34+B15 B35*)
				(*14*)(A11 A14)/Xi^3 + (A12 A24)/Xi^3 + (A13(-A2+A34))/Xi^3 + (A14 A44)/Xi + (A15 A45)/Xi, (*B11 B14+B12 B24+B13 (-A2+B34)+B14 B44+B15 B45*)
				(*15*)(A11 A15)/Xi^3 + (A12 A25)/Xi^3 + (A13 (A1+A35))/Xi^3 + (A14 A45)/Xi + (A15 A55)/Xi, (*B11 B15+B12 B25+B13 (A1+B35)+B14 B54+B15 B55*)
				(*16*)(-A1 A12)/Xi^3 + (A11 A2)/Xi^3 - (A15 A4)/Xi + (A14 A5)/Xi,(*-B1 B12+B11 B2-B15 B4+B14 B5*)
				(*22*) A12^2/Xi^4 + A22^2/Xi^4 + A23^2/Xi^4 +A24^2/Xi^2 +A25^2/Xi^2, (*B21^2+B22^2+B23^2+B24^2+B25^2*)
				(*23*) (A12 A13)/Xi^4 + (A22 A23)/Xi^4 + (A23 A33)/Xi^4 + (A24 A34)/Xi^2 + (A25 A35)/Xi^2, (*B21 B31+B22 B32+B23 B33+B24 B34+B25 B35*)
				(*24*) (A14 A12)/Xi^3 + (A22 A24)/Xi^3 + (A23 (-A2+A34))/Xi^3 + (A24 A44)/Xi + (A25 A45)/Xi, (*B14 B21+B22 B24+B23 (-A2+B34)+B24 B44+B25 B45*)
				(*25*) (A15 A12)/Xi^3 + (A22 A25)/Xi^3 + (A23 (A1+A35))/Xi^3 + (A24 A45)/Xi + (A25 A55)/Xi, (*B15 B21+B22 B25+B23 (B1+B35)+B24 B54+B25 B55*)
				(*26*) (A2 A12)/Xi^3 - (A1 A22)/Xi^3 - (A25 A4)/Xi + (A24 A5)/Xi, (*B2 B21-B1 B22-B25 B4+B24 B5*)
				(*36*) (A2 A13)/Xi^3 - (A1 A23)/Xi^3 - (A35 A4)/Xi + (A34 A5)/Xi,
				(*46*) (A14 A2)/Xi^2 - (A1 A24)/Xi^2 - (A4 A45) + (A44 A5),
				(*56*) (A15 A2)/Xi^2 - (A1 A25)/Xi^2 + (A5 A45) - (A4 A55),
				(*66*) A1^2/Xi^2 + A2^2/Xi^2 + A4^2 + A5^2 (*B1^2+B2^2+B4^2+B5^2*)
			};
			
			(* Return the Components *)
			Return[missingComponents];
			
		]
		
		(*
			External Regularization
		*)
		SpatialComponentsExternalRegularization[orientationList_, spatialComponents_,sigmaSpatialExternal_,sigmaAngularExternal_] := Module[

			{
				R = LieAnalysis`Common`Rotations`Private`RnWithAlphaIs0[orientationList],
				sigmas = ConstantArray[sigmaSpatialExternal,3],
				lMax = LieAnalysis`Common`SphericalHarmonicsLMax[Length@orientationList],
				cartesian, tensor, cartesianTensor, shCoefficients, beta, gamma, spatialComponentsBlurred
			},
			
			(* Construct matrix from components *)
			tensor = {
				{spatialComponents[[1]],spatialComponents[[2]],spatialComponents[[3]]},
				{spatialComponents[[2]],spatialComponents[[4]],spatialComponents[[5]]},
				{spatialComponents[[3]],spatialComponents[[5]],spatialComponents[[6]]}
			}; (* 3,3,x,y,z,R *)
			
			(* Project onto Cartesian Frame *)
			cartesianTensor = Compile[{{tensorIn, _Real, 3}, {RIn, _Real, 3}},
				MapThread[Function[{Rin, tensorC},(#.tensorC.Transpose[#] &)@{
					{Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]]},
				    {Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]]},
				    {Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]]}    
				}], {RIn, tensorIn}]
				, RuntimeAttributes -> {Listable}
			][Transpose[tensor,{5,6,1,2,3,4}], R]; (* x,y,z,R,6,6 *)
			
			(* Apply Spatial Regularization *)
			cartesianTensor = Developer`ToPackedArray@Map[
				GaussianFilter[#,{4sigmas+1,sigmas}]&,
				Transpose[cartesianTensor,{4,5,6,3,1,2}],{3}
			]; (* 3,3,R,x,y,z *)
			
			(* Apply Angular Regularization *)
			shCoefficients = Table[Developer`ToPackedArray@LieAnalysis`Common`RealSphericalHarmonicFit[orientationList,Transpose[cartesianTensor[[i,j]],{4,1,2,3}],lMax] ,{i,1,3},{j,1,3}]; (* 3,3,x,y,z,R*)
			shCoefficients = Developer`ToPackedArray[shCoefficients];
			shCoefficients = LieAnalysis`Common`SphericalHarmonicsAngularBlur[shCoefficients,sigmaAngularExternal];
			
			(* Evaluate SH Coefficients *)
			{beta,gamma} =  Transpose[LieAnalysis`Common`CartesianToSphericalCoordinates[orientationList]][[2;;3]];
			cartesianTensor = shCoefficients.Transpose@(LieAnalysis`Common`SphericalHarmonicsTable[0, lMax][beta,gamma]);
			
			cartesianTensor = Transpose[cartesianTensor,{5,6,1,2,3,4}];
			
			(* Project back to Left Invariant Frame *)
			tensor = Compile[{{tensorIn, _Real, 3}, {RIn, _Real, 3}},
				MapThread[Function[{Rin, tensorC},(Transpose[#].tensorC.# &)@{
					{Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]]},
				    {Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]]},
				    {Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]]} 
				}], {RIn, tensorIn}]
				, RuntimeAttributes -> {Listable}
			][cartesianTensor, R]; (* x,y,z,R,3,3 *)
			
			(* Extract components *)
			spatialComponentsBlurred = {
				tensor[[All,All,All,All,1,1]],
				tensor[[All,All,All,All,1,2]],
				tensor[[All,All,All,All,1,3]],
				tensor[[All,All,All,All,2,2]],
				tensor[[All,All,All,All,2,3]],
				tensor[[All,All,All,All,3,3]]
			};(*6,x,y,z,R*)
			
			(* Return Blurred Spatial Components *)
			Return[spatialComponentsBlurred];
			
		]

		HorizontalComponentsExternalRegularization[orientationList_,spatialFittedOrientationList_, tensorComponents_,sigmaSpatialExternal_,sigmaAngularExternal_] := Module[

			{
				R = LieAnalysis`Common`Rotations`Private`RnWithAlphaIs0[spatialFittedOrientationList],
				sigmas = ConstantArray[sigmaSpatialExternal,3],
				lMax = LieAnalysis`Common`SphericalHarmonicsLMax[Length@orientationList],
				cartesian, tensor, cartesianTensor, shCoefficients, beta, gamma, spatialComponentsBlurred,zero,A11,A12,A13,A14,A15,A16,A22,A23,A24,A25,A26,A36,A46,A56,A66,A33,A34,A35,A44,A45,A55
			},
			
			(* Make Human Readable *)
			{A11,A12,A13,A14,A15,A16,A22,A23,A24,A25,A26,A36,A46,A56,A66,A33,A34,A35,A44,A45,A55} = tensorComponents;
			
			(* Construct matrix from components *)
			zero = ConstantArray[0,Dimensions[A11]];
			tensor = {
				{A11,A12,A13,A14,A15,A16},
				{A12,A22,A23,A24,A25,A26},
				{A13,A23,A33,A34,A35,A36},
				{A14,A24,A34,A44,A45,A46},
				{A15,A25,A35,A45,A55,A56},
				{A16,A26,A36,A46,A56,A66}
			}; (* 6,6,x,y,z,R *)
			
			(* Project onto Cartesian Frame *)
			cartesianTensor = Compile[{{tensorIn, _Real, 3}, {RIn, _Real, 3}},
				MapThread[Function[{Rin, tensorC},(#.tensorC.Transpose[#] &)@{
					{Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]], 0., 0., 0.},
				    {Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]], 0., 0., 0.},
				    {Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]], 0., 0., 0.},
				    {0., 0., 0., Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]]},
				    {0., 0., 0., Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]]},
				    {0., 0., 0., Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]]}
				}], {RIn, tensorIn}]
				, RuntimeAttributes -> {Listable}
			][Transpose[tensor,{5,6,1,2,3,4}], R]; (* x,y,z,R,6,6 *)
			
			(* Apply Spatial Regularization *)
			cartesianTensor = Developer`ToPackedArray@Map[
				GaussianFilter[#,{4sigmas+1,sigmas}]&,
				Transpose[cartesianTensor,{4,5,6,3,1,2}],{3}
			]; (* 3,3,R,x,y,z *)
			
			(* Apply Angular Regularization *)
			shCoefficients = Table[Developer`ToPackedArray@LieAnalysis`Common`RealSphericalHarmonicFit[orientationList,Transpose[cartesianTensor[[i,j]],{4,1,2,3}],lMax] ,{i,1,6},{j,1,6}]; (* 3,3,x,y,z,R*)
			shCoefficients = Developer`ToPackedArray[shCoefficients];
			shCoefficients = LieAnalysis`Common`SphericalHarmonicsAngularBlur[shCoefficients,sigmaAngularExternal];
			
			(* Evaluate SH Coefficients *)
			{beta,gamma} = Transpose[LieAnalysis`Common`CartesianToSphericalCoordinates[orientationList]][[2;;3]];
			cartesianTensor = shCoefficients.Transpose@(LieAnalysis`Common`SphericalHarmonicsTable[0, lMax][beta,gamma]);
			cartesianTensor = Transpose[cartesianTensor,{5,6,1,2,3,4}];
			
			(* Project back to Left Invariant Frame *)
			tensor = Compile[{{tensorIn, _Real, 3}, {RIn, _Real, 3}},
				MapThread[Function[{Rin, tensorC},(Transpose[#].tensorC.# &)@{
					{Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]], 0., 0., 0.},
				    {Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]], 0., 0., 0.},
				    {Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]], 0., 0., 0.},
				    {0., 0., 0., Rin[[1, 1]], Rin[[1, 2]], Rin[[1, 3]]},
				    {0., 0., 0., Rin[[2, 1]], Rin[[2, 2]], Rin[[2, 3]]},
				    {0., 0., 0., Rin[[3, 1]], Rin[[3, 2]], Rin[[3, 3]]}
				}], {RIn, tensorIn}]
				, RuntimeAttributes -> {Listable}
			][cartesianTensor, R]; (* x,y,z,R,3,3 *)
			
			(* Extract components *)
			spatialComponentsBlurred = {
				tensor[[All,All,All,All,3,3]],
				tensor[[All,All,All,All,3,4]],
				tensor[[All,All,All,All,3,5]],
				tensor[[All,All,All,All,4,4]],
				tensor[[All,All,All,All,4,5]],
				tensor[[All,All,All,All,5,5]]
			};(*6,x,y,z,R*)
			
			(* Return Blurred Spatial Components *)
			Return[spatialComponentsBlurred];
			
		]

		(*
			TangentVectorTogaugeFrame
		*)
	
		TangentVectorToGaugeFrame[c_] := TangentVectorToGaugeFrameC[c,1];
		
		
		TangentVectorToGaugeFrame[c_,mu_] := TangentVectorToGaugeFrameC[c,mu];
		
		
		TangentVectorToGaugeFrameR1C=Compile[{{v,_Real,1}},
			Module[{eps=10^-8,v1=v[[1]],v2=v[[2]],v3=v[[3]],c},
				c=v1^2+v2^2;
				If[c>eps,
					{{(v2^2+v1^2 v3)/c,(v1 v2 (-1+v3))/c,v1},{(v1 v2 (-1+v3))/c,(v1^2+v2^2 v3)/c,v2},{-v1,-v2,v3}},
					{{1.`,0.`,0.`},{0.`,1.`,0.`},{0.`,0.`,1.`}}
				]
			] 
		];
		
		
		
		TangentVectorToGaugeFrameR2C=Compile[{{v,_Real,1}},
			Module[{eps=10^-8,v1=v[[1]],v2=v[[2]],v3=v[[3]],v4=v[[4]],c},
				c=v2^2+v3^2+v4^2;
				If[c>eps ,
					{
						{v1,-v2,-v3,-v4},
						{v2,(v1 v2^2+v3^2+v4^2)/c,((-1+v1) v2 v3)/c,((-1+v1) v2 v4)/c},
						{v3,((-1+v1) v2 v3)/c,(v2^2+v1 v3^2+v4^2)/c,((-1+v1) v3 v4)/c},
						{v4,((-1+v1) v2 v4)/c,((-1+v1) v3 v4)/c,(v2^2+v3^2+v1 v4^2)/c}
					},{
						{1.`,0.`,0.`,0.`}
						,{0.`,1.`,0.`,0.`}
						,{0.`,0.`,1.`,0.`}
						,{0.`,0.`,0.`,1.`}
					}
				]
			]
		];
		
		
		
		TangentVectorToGaugeFrameC=Compile[{{c,_Real,1},{mu,_Real}},
			Module[{Rs,Ra,MmuI,Mmu,ctilde},
				Ra=Rs={{1.`,0.`,0.`,0.`,0.`,0.`},{0.`,1.`,0.`,0.`,0.`,0.`},{0.`,0.`,1.`,0.`,0.`,0.`},{0.`,0.`,0.`,1.`,0.`,0.`},{0.`,0.`,0.`,0.`,1.`,0.`},{0.`,0.`,0.`,0.`,0.`,1.`}};
				MmuI={{1/mu,0.`,0.`,0.`,0.`,0.`},{0.`,1/mu,0.`,0.`,0.`,0.`},{0.`,0.`,1/mu,0.`,0.`,0.`},{0.`,0.`,0.`,1.`,0.`,0.`},{0.`,0.`,0.`,0.`,1.`,0.`},{0.`,0.`,0.`,0.`,0.`,1.`}};
				Mmu={{mu,0.`,0.`,0.`,0.`,0.`},{0.`,mu,0.`,0.`,0.`,0.`},{0.`,0.`,mu,0.`,0.`,0.`},{0.`,0.`,0.`,1.`,0.`,0.`},{0.`,0.`,0.`,0.`,1.`,0.`},{0.`,0.`,0.`,0.`,0.`,1.`}};
				ctilde=Mmu.c;
				Rs[[1;;3,1;;3]]=TangentVectorToGaugeFrameR1C[ctilde[[1;;3]]/Norm[ctilde[[1;;3]]]];
				Ra[[3;;6,3;;6]]=TangentVectorToGaugeFrameR2C[{Norm[ctilde[[1;;3]]],ctilde[[4]],ctilde[[5]],ctilde[[6]]}/Norm[ctilde[[1;;6]]]];
		
				(*mu^ is added when we want to follow same convention as in Franken where mu-orthogonality is defined as all vectors having norm equal to \[Mu]*)
				(*mu*) Transpose[Rs.Ra].MmuI
			],
			RuntimeAttributes-> Listable,CompilationOptions -> {"InlineExternalDefinitions" -> True, "InlineCompiledFunctions" -> True}
		];
	
	
	
	End[];



EndPackage[];